
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Xproto.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

Display *disp;
XID xid;
Atom xdel;
XComposeStatus xcompose;
unsigned long last_event_number;
unsigned long last_event_time;
int visible_windows = 0;

void create_window(int x, int y, int width, int height, const char *name)
{
	XSizeHints xsh;
	char *dummy[] = {"app", NULL};

	xid = XCreateSimpleWindow(disp,
		DefaultRootWindow(disp),	/* parent window */
		x,
		y,
		width,
		height,
		0,		/* border width */
		BlackPixel(disp, DefaultScreen(disp)),	/* foreground */
		WhitePixel(disp, DefaultScreen(disp)));	/* background */

	xsh.flags = (PPosition | PSize);
	xsh.x = x;
	xsh.y = y;
	xsh.width = width;
	xsh.height = height;
	XSetStandardProperties(disp,
		xid,	/* the window */
		name,	/* window name */
		name,	/* icon name */
		None,	/* icon pixmap */
		dummy,	/* argv */
		1,	/* argc */
		&xsh	/* size hints */
		);

	XSelectInput(disp, xid,
		(KeyPressMask |
		KeyReleaseMask |
		ButtonPressMask |
		ButtonReleaseMask |
		EnterWindowMask |
		LeaveWindowMask |
		PointerMotionMask |
		ButtonMotionMask |
		KeymapStateMask |
		ExposureMask |
		VisibilityChangeMask |
		StructureNotifyMask |
	/*	ResizeRedirectMask |		*/
	/*	SubstructureNotifyMask |	*/
	/*	SubstructureRedirectMask |	*/
	/*	PropertyChangeMask |		*/
		FocusChangeMask |
		ColormapChangeMask));

	xdel = XInternAtom(disp, "WM_DELETE_WINDOW", False);
	XSetWMProtocols(disp, xid, &xdel, 1);

	XMapRaised(disp, xid);
	visible_windows = 1;
}

#if 0

/*
 *  Convert XK_ X keyboard constants into Unicode characters
 *  or key actions. Unicode characters are sent to the key_down
 *  function of the window, while arrow keys, home, end, etc are
 *  sent to the key_action function. These two are kept separate
 *  because we don't wish to confuse things by just mapping
 *  arrow keys onto Unicode and sending them all to the same
 *  fuction; that would be bad news, because we'd have no way
 *  to distinguish between the user inputting a Unicode char,
 *  and using an arrow key, for instance.
 *
 *  Note: we do map the arrow keys etc onto sensible-seeming
 *  Unicode values, even though they are sent to a different
 *  function. This is incidental and should not be relied upon.
 *  They could have just as easily been mapped onto -1, -2, -3 etc.
 *
 *  The X to Unicode mapping was provided by the folks at XFree86.
 *  Eventually, we can one day hope that Unicode will be the
 *  default key mapping arrangement inside X, and UTF-8 the
 *  default text transfer encoding between X applications, but
 *  until then, we're stuck with the large look-up table found
 *  in keys2ucs.c. It's based on keysym2ucs.c from XFree86,
 *  with the header file removed and function name changed.
 */
static void app_translate_key(Window *win, KeySym key, int shft, int ctrl)
{
	unsigned long value;
	int pass_to = 0; /* 1 is == Unicode key_down, 2 is key_action */

	/* normal keyboard chars get passed to the key_down handler */
	if ((key >= 0x0020 && key <= 0x007e) ||
		(key >= 0x00a0 && key <= 0x00ff))
	{
		value = key;
		pass_to = 1;	/* pass to key_down handler */
	}

	/* other things defined by ASCII get passed there too */
	else if ((key >= XK_BackSpace) && (key <= XK_Escape)) {
		value = (key & 0x00FF); /* mask */
		pass_to = 1;	/* pass to key_down handler */
	}

	/* translate keypad numbers into ASCII digits */
	else if ((key >= XK_KP_0) && (key <= XK_KP_9)) {
		value = key - XK_KP_0 + '0';
		pass_to = 1;	/* pass to key_down handler */
	}

	/* translate other keypad keys into ASCII values */
	else if (key == XK_KP_Space) {
		value = ' ';
		pass_to = 1;	/* pass to key_down handler */
	}
	else if (key == XK_KP_Tab) {
		value = '\t';
		pass_to = 1;	/* pass to key_down handler */
	}
	else if (key == XK_KP_Enter) {
		value = '\n';
		pass_to = 1;	/* pass to key_down handler */
	}
	else if (key == XK_KP_Equal) {
		value = '=';
		pass_to = 1;	/* pass to key_down handler */
	}
	else if (key == XK_KP_Multiply) {
		value = '*';
		pass_to = 1;	/* pass to key_down handler */
	}
	else if (key == XK_KP_Add) {
		value = '+';
		pass_to = 1;	/* pass to key_down handler */
	}
	else if (key == XK_KP_Subtract) {
		value = '-';
		pass_to = 1;	/* pass to key_down handler */
	}
	else if (key == XK_KP_Decimal) {
		value = '.';
		pass_to = 1;	/* pass to key_down handler */
	}
	else if (key == XK_KP_Divide) {
		value = '/';
		pass_to = 1;	/* pass to key_down handler */
	}

	/* translate arrow key combinations into Unicode arrow symbols */
	else if ((key >= XK_Left) && (key <= XK_Down)) {
		value = key - XK_Left + LEFT;
		pass_to = 2;	/* pass to key_action handler */
	}

	/* translate function keys into Unicode circled numbers */
	else if ((key >= XK_F1) && (key <= XK_F10)) {
		value = key - XK_F1 + F1;
		pass_to = 2;	/* pass to key_action handler */
	}
	else if ((key >= XK_KP_F1) && (key <= XK_KP_F4)) {
		value = key - XK_KP_F1 + F1;
		pass_to = 2;	/* pass to key_action handler */
	}

	/* translate other keyboard keys into 'Unicode equivalents' */
	else if (key == XK_Prior) {
		value = PGUP;
		pass_to = 2;	/* pass to key_action handler */
	}
	else if (key == XK_Next) {
		value = PGDN;
		pass_to = 2;	/* pass to key_action handler */
	}
	else if (key == XK_End) {
		value = END;
		pass_to = 2;	/* pass to key_action handler */
	}
	else if (key == XK_Home) {
		value = HOME;
		pass_to = 2;	/* pass to key_action handler */
	}
	else if (key == XK_Insert) {
		value = INS;
		pass_to = 2;	/* pass to key_action handler */
	}
	else if (key == XK_Delete) {
		value = DEL;
		pass_to = 2;	/* pass to key_action handler */
	}

	/* lastly search a table to convert to a Unicode character */
	else {
		value = app_keysym_to_ucs(key); /* see keys2ucs.c */
		if (value != -1)
			pass_to = 1;	/* pass to key_down handler */
	}

	/* now signal Ctrl and Shift key combinations */
	if (ctrl) {
		if (pass_to)
			pass_to = 2;	/* pass to key_action handler */
		value |= CONTROL;
	}
	if ((pass_to == 2) && (shft))
		value |= SHIFT;

	app_send_key_value(win, value, pass_to);
}

#endif


static void do_key_press(XKeyEvent *xkey)
{
	char s[20];
	int i;
	KeySym key;
	int shft, ctrl, alt;

	/* remember Ctrl key clicks, but factor them out */
	shft = (xkey->state & ShiftMask);
	ctrl = (xkey->state & ControlMask);
	alt  = (xkey->state & Mod1Mask);

	/* make key contain an XK_* keycode */
	i = XLookupString(xkey, s, sizeof(s), &key, &xcompose);
	s[i] = '\0';

	printf("window-id=%lu, state=%8.8lx, key=%lx, s=%s\n",
			(unsigned long) xid,
			(unsigned long) xkey->state,
			(unsigned long) key,
			s);
}


/*
 *  This is the function which handles all X-Windows events as
 *  they arrive, and dispatches them to the approprite functions.
 */
static void winproc(XEvent *e)
{
	/* Find window to which this message is directed */
	if (xid != e->xany.window) {
		return; /* not found */
	}

	/* Remember request number in case we need it */
	last_event_number = e->xany.serial;

	switch (e->type)
	{
	  case MotionNotify:
		last_event_time = e->xmotion.time;
		printf("MotionNotify:\twindow-id=%lx, state=%8.8lx, x=%ld, y=%ld\n",
			(unsigned long) xid,
			(unsigned long) e->xmotion.state,
			(long) e->xmotion.x,
			(long) e->xmotion.y);
		return;

	  case ButtonPress:
		last_event_time = e->xbutton.time;
		printf("ButtonPress:\twindow-id=%lx, state=%8.8lx, x=%ld, y=%ld, button=%8.8lx\n",
			(unsigned long) xid,
			(unsigned long) e->xbutton.state,
			(long) e->xbutton.x,
			(long) e->xbutton.y,
			(unsigned long) e->xbutton.button);
		return;

	  case ButtonRelease:
		last_event_time = e->xbutton.time;
		printf("ButtonRelease:\twindow-id=%lx, state=%8.8lx, x=%ld, y=%ld, button=%8.8lx\n",
			(unsigned long) xid,
			(unsigned long) e->xbutton.state,
			(long) e->xbutton.x,
			(long) e->xbutton.y,
			(unsigned long) e->xbutton.button);
		return;

	  case KeyPress:
		last_event_time = e->xkey.time;
		printf("KeyPress:\t");
		do_key_press(&e->xkey);
		return;

	  case KeyRelease:
		last_event_time = e->xkey.time;
		printf("KeyRelease:\t");
		do_key_press(&e->xkey);
		return;

	  case FocusIn:
		printf("FocusIn:\t\n");
		return;

	  case FocusOut:
		printf("FocusOut:\t\n");
		return;

	  case MappingNotify:
		XRefreshKeyboardMapping(&e->xmapping);
		printf("MappingNotify:\t\n");
		return;

	  case ResizeRequest:
		printf("ResizeRequest:\twindow-id=%lx, width=%ld, height=%ld\n",
			(unsigned long) xid,
			(long) e->xresizerequest.width,
			(long) e->xresizerequest.height);
		return;

	  case ConfigureNotify:
		printf("ConfigureNotify:\twindow-id=%lx, x=%ld, y=%ld, width=%ld, height=%ld\n",
			(unsigned long) xid,
			(long) e->xconfigurerequest.x,
			(long) e->xconfigurerequest.y,
			(long) e->xconfigurerequest.width,
			(long) e->xconfigurerequest.height);
		return;

	  case GraphicsExpose:
		printf("Graphics");
	  case Expose:
		printf("Expose:\twindow-id=%lx, x=%ld, y=%ld, width=%ld, height=%ld\n",
			(unsigned long) xid,
			(long) e->xexpose.x,
			(long) e->xexpose.y,
			(long) e->xexpose.width,
			(long) e->xexpose.height);
		return;

	  case NoExpose:
		printf("NoExpose:\t\n");
		return;

	  case ClientMessage:
		printf("ClientMessage:\t");
		if (((Atom *) &e->xclient.data)[0] == xdel) {
			printf("(close-window)\n");
			exit(0);
		}
		printf("(other)\n");
		return;

	  case DestroyNotify:
		printf("DestroyNotify:\t (close-window)\n");
		exit(0);
		return;

	  case SelectionRequest:
		last_event_time = e->xselectionrequest.time;
		printf("SelectionRequest:\t\n");
		return;

	  case SelectionClear:
		last_event_time = e->xselectionclear.time;
		printf("SelectionClear:\t\n");
		return;

	  case SelectionNotify:
		last_event_time = e->xselection.time;
		printf("SelectionNotify:\t\n");
		return;

	  default:
		printf("%lu:\t\n", (unsigned long) e->type);
		break;
	}
}


int wait_event()
{
	XEvent event;
	int result = 0;

	if (visible_windows > 0) {
		XNextEvent(disp, &event);
		result = 1;
		winproc(&event);
	}

	return result;
}

/*
 *  Handle all events until there are no more windows:
 */
void main_loop()
{
	while (wait_event())
		continue;
}

void draw_all()
{
	XSync(disp, 0);
}

int main(int argc, char *argv[])
{
	char *display_name;

	display_name = getenv("DISPLAY");
	if (display_name == NULL) {
		return 1;
	}

	/* The GUI should be available, excellent */
	//XSetErrorHandler(app_x_error_handler);
	//XSetIOErrorHandler(app_x_io_error_handler);
	disp = XOpenDisplay(display_name);

	create_window(100, 100, 400, 200, "X11 test");
	main_loop();
	XCloseDisplay(disp);
	return 0;
}

